@@ -204,11 +204,17 @@ class SourceWalker(GenericASTTraversal, object):
                 'raise_stmt2':	( '%|raise %c, %c\n', 0, 1),
             })
         else:
-            # Gotta love Python for its futzing around with syntax like this
             TABLE_DIRECT.update({
-                'raise_stmt2':	( '%|raise %c from %c\n', 0, 1),
+                # Gotta love Python for its futzing around with syntax like this
+                'raise_stmt2':	 ( '%|raise %c from %c\n', 0, 1),
             })
 
+            if version >= 3.2:
+                TABLE_DIRECT.update({
+                'del_deref_stmt': ( '%|del %c\n', 0),
+                'DELETE_DEREF': ( '%{pattr}', 0 ),
+                })
+
         if version < 2.0:
             TABLE_DIRECT.update({
                 'importlist':	( '%C', (0, maxint, ', ') ),
@@ -218,6 +224,11 @@ class SourceWalker(GenericASTTraversal, object):
                 'importlist2':	( '%C', (0, maxint, ', ') ),
                 })
             if version <= 2.4:
+                if version == 2.3:
+                    TABLE_DIRECT.update({
+                        'if1_stmt':	( '%|if 1\n%+%c%-', 5 )
+                    })
+
                 global NAME_MODULE
                 NAME_MODULE = AST('stmt',
                                   [ AST('assign',
@@ -324,7 +335,8 @@ class SourceWalker(GenericASTTraversal, object):
                 #######################
                 if version >= 3.5:
                     TABLE_DIRECT.update({
-                        'await_stmt':	       ( '%|await %c', 0),
+                        'await_expr':	       ( 'await %c', 0),
+                        'await_stmt':	       ( '%|%c', 0),
                         'async_for_stmt':      (
                             '%|async for %c in %c:\n%+%c%-\n\n', 9, 1, 25 ),
                         'async_forelse_stmt':  (
@@ -333,6 +345,8 @@ class SourceWalker(GenericASTTraversal, object):
                             '%|async with %c:\n%+%c%-', 0, 7),
                         'async_with_as_stmt':  (
                             '%|async with %c as %c:\n%+%c%-', 0, 6, 7),
+                        'unmap_dict':	       ( '{**%C}', (0, -1, ', **') ),
+                        # 'unmapexpr':	       ( '{**%c}', 0), # done by n_unmapexpr
 
                     })
                     def n_async_call_function(node):
@@ -344,11 +358,43 @@ class SourceWalker(GenericASTTraversal, object):
                         self.prec = p
                         self.prune()
                     self.n_async_call_function = n_async_call_function
+                    self.n_build_list_unpack = self.n_build_list
+
+                    if version == 3.5:
+                        def n_call_function(node):
+                            mapping = self._get_mapping(node)
+                            table = mapping[0]
+                            key = node
+                            for i in mapping[1:]:
+                                key = key[i]
+                                pass
+                            if key.type.startswith('CALL_FUNCTION_VAR_KW'):
+                                # Python 3.5 changes the stack position of *args. kwargs come
+                                # after *args whereas in earlier Pythons, *args is at the end
+                                # which simpilfiies things from our perspective.
+                                # Python 3.6+ replaces CALL_FUNCTION_VAR_KW with CALL_FUNCTION_EX
+                                # We will just swap the order to make it look like earlier Python 3.
+                                entry = table[key.type]
+                                kwarg_pos = entry[2][1]
+                                args_pos = kwarg_pos - 1
+                                # Put last node[args_pos] after subsequent kwargs
+                                while node[kwarg_pos] == 'kwarg' and kwarg_pos < len(node):
+                                    # swap node[args_pos] with node[kwargs_pos]
+                                    node[kwarg_pos], node[args_pos] = node[args_pos], node[kwarg_pos]
+                                    args_pos = kwarg_pos
+                                    kwarg_pos += 1
+                            self.default(node)
+                        self.n_call_function = n_call_function
 
                     def n_funcdef(node):
-                        code_node = node[0][1]
-                        if (code_node == 'LOAD_CONST' and iscode(code_node.attr)
-                            and code_node.attr.co_flags & COMPILER_FLAG_BIT['COROUTINE']):
+                        if self.version == 3.6:
+                            code_node = node[0][0]
+                        else:
+                            code_node = node[0][1]
+
+                        is_code = hasattr(code_node, 'attr') and iscode(code_node.attr)
+                        if (is_code and
+                            (code_node.attr.co_flags & COMPILER_FLAG_BIT['COROUTINE'])):
                             self.engine(('\n\n%|async def %c\n', -2), node)
                         else:
                             self.engine(('\n\n%|def %c\n', -2), node)
@@ -363,8 +409,6 @@ class SourceWalker(GenericASTTraversal, object):
                                 self.f.write(', **')
                                 pass
                             pass
-                        if version >= 3.6:
-                            self.f.write(')')
                         self.prune()
                         pass
                     self.n_unmapexpr = n_unmapexpr
@@ -378,8 +422,13 @@ class SourceWalker(GenericASTTraversal, object):
                         'fstring_single': ( "f'{%c%{conversion}}'", 0),
                         'fstring_multi':  ( "f'%c'", 0),
                         'func_args36':    ( "%c(**", 0),
+                        #'kwargs_only_36': ( "%c(**", 0),
                     })
-
+                    TABLE_R.update({
+                        'CALL_FUNCTION_EX': ('%c(*%P)', 0, (1, 2, ', ', 100)),
+                        # Not quite right
+                        'CALL_FUNCTION_EX_KW': ('%c(**%C)', 0, (2,3, ',')),
+                        })
                     FSTRING_CONVERSION_MAP = {1: '!s', 2: '!r', 3: '!a'}
 
                     def f_conversion(node):
@@ -395,6 +444,27 @@ class SourceWalker(GenericASTTraversal, object):
                         self.default(node)
                     self.n_fstring_single = n_fstring_single
 
+                    def n_kwargs_only_36(node):
+                        keys = node[-1].attr
+                        num_kwargs = len(keys)
+                        values = node[:num_kwargs]
+                        for i, (key, value) in enumerate(zip(keys, values)):
+                            self.write(key + '=')
+                            self.preorder(value)
+                            if i < num_kwargs:
+                                self.write(',')
+                        self.prune()
+                        return
+                    self.n_kwargs_only_36 = n_kwargs_only_36
+
+                    def n_return_closure(node):
+                        # Nothing should be output here
+                        self.prune()
+                        return
+                    self.n_return_closure = n_return_closure
+                    pass # version > 3.6
+                pass # version > 3.4
+            pass # version > 3.0
         return
 
     f = property(lambda s: s.params['f'],
@@ -501,6 +571,33 @@ class SourceWalker(GenericASTTraversal, object):
                     node == AST('return_stmt',
                                 [AST('ret_expr', [NONE]), Token('RETURN_VALUE')]))
 
+    ## The below doesn't work because continue may be the only thing inside an 'else'. For example
+    # for ...
+    #   if ...
+    #   else:
+    #     continue
+    #
+    # def n_continue_stmt(self, node):
+    #     if self.version >= 3.0 and node[0] == 'CONTINUE':
+    #         t = node[0]
+    #         if not t.linestart:
+    #             # Artificially-added "continue" statements derived from JUMP_ABSOLUTE
+    #             # don't have line numbers associated with them.
+    #             # If this is a CONTINUE is to the same target as a JUMP_ABSOLUTE following it,
+    #             # then the "continue" can be suppressed.
+    #             op, offset = t.op, t.offset
+    #             next_offset = self.scanner.next_offset(op, offset)
+    #             scanner = self.scanner
+    #             code = scanner.code
+    #             if next_offset < len(code):
+    #                 next_inst = code[next_offset]
+    #                 if (scanner.opc.opname[next_inst] == 'JUMP_ABSOLUTE'
+    #                     and t.pattr == code[next_offset+1]):
+    #                     # Suppress "continue"
+    #                     import pdb; pdb.set_trace()
+    #                     self.prune()
+    #     self.default(node)
+
     def n_return_stmt(self, node):
         if self.params['isLambda']:
             self.preorder(node[0])
@@ -834,19 +931,19 @@ class SourceWalker(GenericASTTraversal, object):
             # LOAD_CONST code object ..
             # LOAD_CONST        'x0'  if >= 3.3
             # MAKE_FUNCTION ..
-            code = node[-3]
+            code_node = node[-3]
         elif node[-2] == 'expr':
-            code = node[-2][0]
+            code_node = node[-2][0]
         else:
             # LOAD_CONST code object ..
             # MAKE_FUNCTION ..
-            code = node[-2]
+            code_node = node[-2]
 
-        func_name = code.attr.co_name
+        func_name = code_node.attr.co_name
         self.write(func_name)
 
         self.indentMore()
-        self.make_function(node, isLambda=False, codeNode=code)
+        self.make_function(node, isLambda=False, codeNode=code_node)
 
         if len(self.param_stack) > 1:
             self.write('\n\n')
@@ -1238,11 +1335,18 @@ class SourceWalker(GenericASTTraversal, object):
 
         if self.version > 3.0:
             if node == 'classdefdeco2':
-                currentclass = node[1][2].pattr
+                if self.version >= 3.6:
+                    class_name = node[1][1].pattr
+                else:
+                    class_name = node[1][2].pattr
                 buildclass = node
             else:
-                currentclass = node[1][0].pattr
-                buildclass = node[0]
+                if self.version >= 3.6:
+                    class_name = node[0][1][0].attr.co_name
+                    buildclass = node[0]
+                else:
+                    class_name = node[1][0].pattr
+                    buildclass = node[0]
 
             assert 'mkfunc' == buildclass[1]
             mkfunc = buildclass[1]
@@ -1250,16 +1354,16 @@ class SourceWalker(GenericASTTraversal, object):
                 if 3.0 <= self.version <= 3.2:
                     for n in mkfunc:
                         if hasattr(n, 'attr') and iscode(n.attr):
-                            subclass = n.attr
+                            subclass_code = n.attr
                             break
                         elif n == 'expr':
-                            subclass = n[0].attr
+                            subclass_code = n[0].attr
                         pass
                     pass
                 else:
                     for n in mkfunc:
                         if hasattr(n, 'attr') and iscode(n.attr):
-                            subclass = n.attr
+                            subclass_code = n.attr
                             break
                         pass
                     pass
@@ -1271,10 +1375,10 @@ class SourceWalker(GenericASTTraversal, object):
                     # Python 3.3 classes with closures work like this.
                     # Note have to test before 3.2 case because
                     # index -2 also has an attr.
-                    subclass = load_closure[-3].attr
+                    subclass_code = load_closure[-3].attr
                 elif hasattr(load_closure[-2], 'attr'):
                     # Python 3.2 works like this
-                    subclass = load_closure[-2].attr
+                    subclass_code = load_closure[-2].attr
                 else:
                     raise 'Internal Error n_classdef: cannot find class body'
                 if hasattr(buildclass[3], '__len__'):
@@ -1283,18 +1387,21 @@ class SourceWalker(GenericASTTraversal, object):
                     subclass_info = buildclass[2]
                 else:
                     raise 'Internal Error n_classdef: cannot superclass name'
+            elif self.version >= 3.6 and node == 'classdefdeco2':
+                subclass_info = node
+                subclass_code = buildclass[1][0].attr
             else:
-                subclass = buildclass[1][0].attr
+                subclass_code = buildclass[1][0].attr
                 subclass_info = node[0]
         else:
             buildclass = node if (node == 'classdefdeco2') else node[0]
             build_list = buildclass[1][0]
             if hasattr(buildclass[-3][0], 'attr'):
-                subclass = buildclass[-3][0].attr
-                currentclass = buildclass[0].pattr
+                subclass_code = buildclass[-3][0].attr
+                class_name = buildclass[0].pattr
             elif hasattr(node[0][0], 'pattr'):
-                subclass = buildclass[-3][1].attr
-                currentclass = node[0][0].pattr
+                subclass_code = buildclass[-3][1].attr
+                class_name = node[0][0].pattr
             else:
                 raise 'Internal Error n_classdef: cannot find class name'
 
@@ -1303,7 +1410,7 @@ class SourceWalker(GenericASTTraversal, object):
         else:
             self.write('\n\n')
 
-        self.currentclass = str(currentclass)
+        self.currentclass = str(class_name)
         self.write(self.indent, 'class ', self.currentclass)
 
         if self.version > 3.0:
@@ -1314,7 +1421,7 @@ class SourceWalker(GenericASTTraversal, object):
 
         # class body
         self.indentMore()
-        self.build_class(subclass)
+        self.build_class(subclass_code)
         self.indentLess()
 
         self.currentclass = cclass
@@ -1415,7 +1522,7 @@ class SourceWalker(GenericASTTraversal, object):
                     i += 2
                     pass
                 pass
-            elif node[1].type.startswith('kvlist'):
+            elif len(node) > 1 and node[1].type.startswith('kvlist'):
                 # Python 3.0..3.4 style key/value list in mapexpr
                 kv_node = node[1]
                 l = list(kv_node)
@@ -1450,10 +1557,21 @@ class SourceWalker(GenericASTTraversal, object):
                 values = node[:-2]
                 # FIXME: Line numbers?
                 for key, value in zip(keys, values):
+                    self.write(sep)
                     self.write(repr(key))
+                    line_number = self.line_number
                     self.write(':')
                     self.write(self.traverse(value[0]))
-                    self.write(',')
+                    sep = ","
+                    if line_number != self.line_number:
+                        sep += "\n" + self.indent + INDENT_PER_LEVEL[:-1]
+                        line_number = self.line_number
+                    else:
+                        sep += " "
+                        pass
+                    pass
+                if sep.startswith(",\n"):
+                    self.write(sep[1:])
                 pass
             pass
         else:
@@ -1507,6 +1625,9 @@ class SourceWalker(GenericASTTraversal, object):
                 if line_number != self.line_number:
                     sep += "\n" + self.indent + "  "
                     line_number = self.line_number
+                    pass
+                pass
+            pass
         if sep.startswith(",\n"):
             self.write(sep[1:])
         self.write('}')
@@ -1529,22 +1650,25 @@ class SourceWalker(GenericASTTraversal, object):
         # will assume that if the text ends in *.
         last_was_star = self.f.getvalue().endswith('*')
 
-        have_star = False
         if lastnodetype.endswith('UNPACK'):
             # FIXME: need to handle range of BUILD_LIST_UNPACK
             have_star = True
-            endchar = ''
+            # endchar = ''
         else:
-            if lastnodetype.startswith('BUILD_LIST'):
-                self.write('['); endchar = ']'
-            elif lastnodetype.startswith('BUILD_TUPLE'):
-                self.write('('); endchar = ')'
-            elif lastnodetype.startswith('BUILD_SET'):
-                self.write('{'); endchar = '}'
-            elif lastnodetype.startswith('ROT_TWO'):
-                self.write('('); endchar = ')'
-            else:
-                raise 'Internal Error: n_build_list expects list or tuple'
+            have_star = False
+
+        if lastnodetype.startswith('BUILD_LIST'):
+            self.write('['); endchar = ']'
+        elif lastnodetype.startswith('BUILD_TUPLE'):
+            self.write('('); endchar = ')'
+        elif lastnodetype.startswith('BUILD_SET'):
+            self.write('{'); endchar = '}'
+        elif lastnodetype.startswith('BUILD_MAP_UNPACK'):
+            self.write('{*'); endchar = '}'
+        elif lastnodetype.startswith('ROT_TWO'):
+            self.write('('); endchar = ')'
+        else:
+            raise TypeError('Internal Error: n_build_list expects list, tuple, set, or unpack')
 
         flat_elems = []
         for elem in node:
@@ -1562,7 +1686,7 @@ class SourceWalker(GenericASTTraversal, object):
         sep = ''
 
         for elem in flat_elems:
-            if elem == 'ROT_THREE':
+            if elem in ('ROT_THREE', 'EXTENDED_ARG'):
                 continue
             assert elem == 'expr'
             line_number = self.line_number
@@ -1651,12 +1775,8 @@ class SourceWalker(GenericASTTraversal, object):
 
             typ = m.group('type') or '{'
             node = startnode
-            try:
-                if m.group('child'):
-                    node = node[int(m.group('child'))]
-            except:
-                print(node.__dict__)
-                raise
+            if m.group('child'):
+                node = node[int(m.group('child'))]
 
             if   typ == '%':	self.write('%')
             elif typ == '+':
@@ -1693,6 +1813,8 @@ class SourceWalker(GenericASTTraversal, object):
                     remaining -= 1
                     if remaining > 0:
                         self.write(sep)
+                        pass
+                    pass
                 arg += 1
             elif typ == 'D':
                 low, high, sep = entry[arg]
@@ -1729,7 +1851,6 @@ class SourceWalker(GenericASTTraversal, object):
                 try:
                     self.write(eval(expr, d, d))
                 except:
-                    print(node)
                     raise
             m = escape.search(fmt, i)
         self.write(fmt[i:])
@@ -1760,6 +1881,8 @@ class SourceWalker(GenericASTTraversal, object):
             if k.startswith('CALL_METHOD'):
                 # This happens in PyPy only
                 TABLE_R[k] = ('%c(%P)', 0, (1, -1, ', ', 100))
+            elif self.version >= 3.6 and k.startswith('CALL_FUNCTION_KW'):
+                TABLE_R[k] = ('%c(%P)', 0, (1, -1, ', ', 100))
             elif op == 'CALL_FUNCTION':
                 TABLE_R[k] = ('%c(%P)', 0, (1, -1, ', ', 100))
             elif op in ('CALL_FUNCTION_VAR',
@@ -1771,16 +1894,35 @@ class SourceWalker(GenericASTTraversal, object):
                     str = '%c(%C, '
                     p2 = (1, -2, ', ')
                 if op == 'CALL_FUNCTION_VAR':
-                    str += '*%c)'
+                    # Python 3.5 only puts optional args (the VAR part)
+                    # lowest down the stack
+                    if self.version == 3.5:
+                        if str == '%c(%C, ':
+                            str = '%c(*%C, %c)'
+                        elif str == '%c(%C':
+                            str = '%c(*%C)'
+                            # p2 = (1, -1, 100)
+                    else:
+                        str += '*%c)'
                     entry = (str, 0, p2, -2)
                 elif op == 'CALL_FUNCTION_KW':
                     str += '**%c)'
                     entry = (str, 0, p2, -2)
-                else:
+                elif op == 'CALL_FUNCTION_VAR_KW':
                     str += '*%c, **%c)'
-                    if p2[2]: p2 = (1, -3, ', ')
-                    entry = (str, 0, p2, -3, -2)
+                    # Python 3.5 only puts optional args (the VAR part)
+                    # lowest down the stack
+                    na = (v & 0xff)  # positional parameters
+                    if self.version == 3.5 and na == 0:
+                        if p2[2]: p2 = (2, -2, ', ')
+                        entry = (str, 0, p2, 1, -2)
+                    else:
+                        if p2[2]: p2 = (1, -3, ', ')
+                        entry = (str, 0, p2, -3, -2)
                     pass
+                else:
+                    assert False, "Unhandled CALL_FUNCTION %s" % op
+
                 TABLE_R[k] = entry
                 pass
             # handled by n_mapexpr:
